/***************************************************************************
 *   Copyright (C) 2005-2006 by                                            *
 *                                                                         *
 *      Stefan Schwarzer <sxs@morphine.tv>                                 *
 *      Guido Madaus     <bere@morphine.tv>                                *
 *      Jens Schneider   <pupeider@morphine.tv>                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License.        *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "mmsepgimportplugin.h"
#include <errno.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
									
#include <linux/dvb/dmx.h>
#include "si_tables.h"


MMS_EXPORT_IMPORT_PLUGIN(MMSEPGImportPlugin);

MMSEPGImportPlugin::MMSEPGImportPlugin() {
    this->dump = false;
    this->dumpfile = NULL;

    /* get time offset */
    time_t now = time(NULL);
    struct tm tm;
    localtime_r(&now,&tm);
    this->time_offset = tm.tm_gmtoff / 3600;
}
      
MMSEPGImportPlugin::~MMSEPGImportPlugin() {
}

bool MMSEPGImportPlugin::initialize(MMSPluginData plugindata) {
    MMSConfigData   *config;
    DataSource      confsource;
    DataSource      datasource;

    logger.setIdentity("MMSEPGImportPlugin");

    /* build data source */
    config = new MMSConfigData();
    confsource.setFilename(config->getConfigDBName());
    datasource.setFilename(config->getDataDBName());
    delete config;

    /* save my plugin data */
    this->plugindata = plugindata;

    /* through the plugin properties */
    vector<MMSPropertyData*> probs = this->plugindata.getProperties();
    for (unsigned int i = 0; i < probs.size(); i++) {
        if (probs.at(i)->getParameter()=="DUMP") {
            /* should I dump the stream? */
            this->dump = (probs.at(i)->getValue()=="YES");
        }
    }

    /* get access to source service (config data) */
    iss = new MMSImportSourceService(confsource);
    
    /* get access to tvchannel service (my data) */
    tcs = new MMSTVChannelService(datasource);

    /* get access to epg service (my data) */
    eps = new MMSEPGService(datasource);
    
    return true;
}


bool MMSEPGImportPlugin::execute() {
    vector<MMSImportSourceData *>   sourcesList;
    vector<MMSTVChannelData *>      channels;

    logger.writeLog("IMPORTING...");

    try {
        /* get all import sources */
        sourcesList = iss->getImportSourcesByPlugin(&(this->plugindata));

        /* get channels */
        channels = tcs->getAllChannels();         
    
        /* through all sources */
        for (unsigned i=0;i<sourcesList.size();i++) {
            /* work with source */
            logger.writeLog("Read EPG from stream " + sourcesList.at(i)->getSource());

            /* epg data */
            vector<MMSEPGData *> epg;

            /* read epg data to channels */
            if (!readEPG(channels, epg, sourcesList.at(i)->getSource()))
                continue;

            /* work with epg */
            logger.writeLog("Synchronize EPG with database");
            eps->setEPGData(epg);            

            /* free memory */
            for (unsigned i=0;i<epg.size();i++) {
                delete epg.at(i);
                epg.erase(epg.begin()+i);
            }
        }
    
        /* free memory */
        for (unsigned i=0;i<channels.size();i++)
            delete channels.at(i);

        /* free memory */
        for (unsigned i=0;i<sourcesList.size();i++)
            delete sourcesList.at(i);

    } catch(MMSError *error) {
        logger.writeLog("Abort due to: " + error->getMessage());
    }

    return true;
}

bool MMSEPGImportPlugin::shutdown() {
    delete iss;
    delete tcs;
    delete eps;
    return true;
}

bool MMSEPGImportPlugin::cleanUp() {
    return true;
}


static uint32_t crc_table[256] = {
	0x00000000, 0x04c11db7, 0x09823b6e, 0x0d4326d9, 0x130476dc, 0x17c56b6b,
    0x1a864db2, 0x1e475005, 0x2608edb8, 0x22c9f00f, 0x2f8ad6d6, 0x2b4bcb61,
    0x350c9b64, 0x31cd86d3, 0x3c8ea00a, 0x384fbdbd, 0x4c11db70, 0x48d0c6c7,
    0x4593e01e, 0x4152fda9, 0x5f15adac, 0x5bd4b01b, 0x569796c2, 0x52568b75,
    0x6a1936c8, 0x6ed82b7f, 0x639b0da6, 0x675a1011, 0x791d4014, 0x7ddc5da3,
    0x709f7b7a, 0x745e66cd, 0x9823b6e0, 0x9ce2ab57, 0x91a18d8e, 0x95609039,
    0x8b27c03c, 0x8fe6dd8b, 0x82a5fb52, 0x8664e6e5, 0xbe2b5b58, 0xbaea46ef,
    0xb7a96036, 0xb3687d81, 0xad2f2d84, 0xa9ee3033, 0xa4ad16ea, 0xa06c0b5d,
    0xd4326d90, 0xd0f37027, 0xddb056fe, 0xd9714b49, 0xc7361b4c, 0xc3f706fb,
    0xceb42022, 0xca753d95, 0xf23a8028, 0xf6fb9d9f, 0xfbb8bb46, 0xff79a6f1,
    0xe13ef6f4, 0xe5ffeb43, 0xe8bccd9a, 0xec7dd02d, 0x34867077, 0x30476dc0,
    0x3d044b19, 0x39c556ae, 0x278206ab, 0x23431b1c, 0x2e003dc5, 0x2ac12072,
    0x128e9dcf, 0x164f8078, 0x1b0ca6a1, 0x1fcdbb16, 0x018aeb13, 0x054bf6a4,
    0x0808d07d, 0x0cc9cdca, 0x7897ab07, 0x7c56b6b0, 0x71159069, 0x75d48dde,
    0x6b93dddb, 0x6f52c06c, 0x6211e6b5, 0x66d0fb02, 0x5e9f46bf, 0x5a5e5b08,
    0x571d7dd1, 0x53dc6066, 0x4d9b3063, 0x495a2dd4, 0x44190b0d, 0x40d816ba,
    0xaca5c697, 0xa864db20, 0xa527fdf9, 0xa1e6e04e, 0xbfa1b04b, 0xbb60adfc,
    0xb6238b25, 0xb2e29692, 0x8aad2b2f, 0x8e6c3698, 0x832f1041, 0x87ee0df6,
    0x99a95df3, 0x9d684044, 0x902b669d, 0x94ea7b2a, 0xe0b41de7, 0xe4750050,
    0xe9362689, 0xedf73b3e, 0xf3b06b3b, 0xf771768c, 0xfa325055, 0xfef34de2,
    0xc6bcf05f, 0xc27dede8, 0xcf3ecb31, 0xcbffd686, 0xd5b88683, 0xd1799b34,
    0xdc3abded, 0xd8fba05a, 0x690ce0ee, 0x6dcdfd59, 0x608edb80, 0x644fc637,
    0x7a089632, 0x7ec98b85, 0x738aad5c, 0x774bb0eb, 0x4f040d56, 0x4bc510e1,
    0x46863638, 0x42472b8f, 0x5c007b8a, 0x58c1663d, 0x558240e4, 0x51435d53,
    0x251d3b9e, 0x21dc2629, 0x2c9f00f0, 0x285e1d47, 0x36194d42, 0x32d850f5,
    0x3f9b762c, 0x3b5a6b9b, 0x0315d626, 0x07d4cb91, 0x0a97ed48, 0x0e56f0ff,
    0x1011a0fa, 0x14d0bd4d, 0x19939b94, 0x1d528623, 0xf12f560e, 0xf5ee4bb9,
    0xf8ad6d60, 0xfc6c70d7, 0xe22b20d2, 0xe6ea3d65, 0xeba91bbc, 0xef68060b,
    0xd727bbb6, 0xd3e6a601, 0xdea580d8, 0xda649d6f, 0xc423cd6a, 0xc0e2d0dd,
    0xcda1f604, 0xc960ebb3, 0xbd3e8d7e, 0xb9ff90c9, 0xb4bcb610, 0xb07daba7,
    0xae3afba2, 0xaafbe615, 0xa7b8c0cc, 0xa379dd7b, 0x9b3660c6, 0x9ff77d71,
    0x92b45ba8, 0x9675461f, 0x8832161a, 0x8cf30bad, 0x81b02d74, 0x857130c3,
    0x5d8a9099, 0x594b8d2e, 0x5408abf7, 0x50c9b640, 0x4e8ee645, 0x4a4ffbf2,
    0x470cdd2b, 0x43cdc09c, 0x7b827d21, 0x7f436096, 0x7200464f, 0x76c15bf8,
    0x68860bfd, 0x6c47164a, 0x61043093, 0x65c52d24, 0x119b4be9, 0x155a565e,
    0x18197087, 0x1cd86d30, 0x029f3d35, 0x065e2082, 0x0b1d065b, 0x0fdc1bec,
    0x3793a651, 0x3352bbe6, 0x3e119d3f, 0x3ad08088, 0x2497d08d, 0x2056cd3a,
    0x2d15ebe3, 0x29d4f654, 0xc5a92679, 0xc1683bce, 0xcc2b1d17, 0xc8ea00a0,
    0xd6ad50a5, 0xd26c4d12, 0xdf2f6bcb, 0xdbee767c, 0xe3a1cbc1, 0xe760d676,
    0xea23f0af, 0xeee2ed18, 0xf0a5bd1d, 0xf464a0aa, 0xf9278673, 0xfde69bc4,
    0x89b8fd09, 0x8d79e0be, 0x803ac667, 0x84fbdbd0, 0x9abc8bd5, 0x9e7d9662,
    0x933eb0bb, 0x97ffad0c, 0xafb010b1, 0xab710d06, 0xa6322bdf, 0xa2f33668,
	0xbcb4666d, 0xb8757bda, 0xb5365d03, 0xb1f740b4
};

uint32_t _dvb_crc32(const uint8_t *data, size_t len) {
    uint32_t crc = 0xffffffff;
    for (unsigned int i = 0; i < len; i++)
		crc = (crc << 8) ^ crc_table[((crc >> 24) ^ *data++) & 0xff];
	return crc;
}

#define bcdtoint(i) ((((i & 0xf0) >> 4) * 10) + (i & 0x0f))


/* Datastructure to parse ******************/
/*                                         */
/* abcccd...e...                           */
/*                                         */
/* a   - tag/id                            */ 
/* b   - number of following bytes for tag */ 
/* ccc - language                          */
/* d   - length of title                   */
/* ... - title                             */
/* e   - length of description             */
/* ... - description                       */
/*                                         */
/* Example for three bytes title and four  */
/* bytes description:                      */
/* 4D0C------03------04--------            */
/*******************************************/
void MMSEPGImportPlugin::parseShortEventDesc(char *evtdesc, MMSEPGData *data) 
{
    int evtlen,dsclen;
    char lang[4];
    char evt[256];
    char dsc[256];
    evtlen=(*(evtdesc+5))&0xff;
    dsclen=(*(evtdesc+6+evtlen))&0xff;
    strncpy(lang,(char*)(evtdesc+2),3);
    strncpy(evt,(char*)(evtdesc+6),evtlen);
    strncpy(dsc,(char*)(evtdesc+7+evtlen),dsclen);
    lang[3]=0;
    evt[evtlen]=0;
    dsc[dsclen]=0;

    writeDumpStr("LANG = ");
    writeDumpStr(lang);
    writeDumpStr("\nEVT = ");
    writeDumpStr(evt);
    writeDumpStr("\nDSC = ");
    writeDumpStr(dsc);
    writeDumpStr("\n");

    if (*evt>=' ')
        data->setTitle(data->getTitle() + evt);
    else
    if (*evt>0)
        data->setTitle(data->getTitle() + &evt[1]);

    if (*dsc>=' ')
        data->setDescription(data->getDescription() + dsc);
    else   
    if (*dsc>0)
        data->setDescription(data->getDescription() + &dsc[1]);
}

/* Datastructure to parse ******************/
/*                                         */
/* abcdddef...g...f...g...h...             */
/*                                         */
/* a   - tag/id                            */ 
/* b   - number of following bytes for tag */
/* c   - other numbers each 4bit           */ 
/* ddd - language                          */
/* e   - length of item block              */
/*       the item block can have more than */
/*       one first/second part pair        */
/* f   - lenght of first part of the item  */
/* ... - first part                        */
/* g   - length of second part of the item */
/* ... - second part                       */
/* h   - length of the text                */
/* ... - text                              */
/*                                         */
/* Example for only four bytes text:       */
/* 4E0C00------0004--------                */
/*******************************************/
void MMSEPGImportPlugin::parseLongEventDesc(char *evtdesc, MMSEPGData *data) {
    string fs="";

    int firstsize = evtdesc[6];
    for (int j = 0; j < firstsize;) {
        int item_description_length = evtdesc[7+j];
        char s[256];
        if (item_description_length) {
        	memcpy(s,&evtdesc[7+j+1], item_description_length);
        	s[item_description_length]=0;

            writeDumpStr("ITEMDESC = ");
            writeDumpStr(s);
            writeDumpStr("\n");

            if (item_description_length) {
                if (fs != "")
                    fs+=", ";

                if (*s>=' ')
                    fs+= s;
                else   
                if (*s>0)
                    fs+= &s[1];
            }
        }
        j+=item_description_length+1;
        
        int item_length = evtdesc[7+j];
        if (item_length) {
        	memcpy(s,&evtdesc[7+j+1], item_length);
        	s[item_length]=0;

            writeDumpStr("ITEM = ");
            writeDumpStr(s);
            writeDumpStr("\n");

            if (item_length) {
                if (fs != "")
                    fs+=", ";

                if (*s>=' ')
                    fs+= s;
                else   
                if (*s>0)
                    fs+= &s[1];
            }
        }
        j+=item_length+1;
    }

    int g=(unsigned char)evtdesc[7+firstsize];
    if (g>0) {
        char s[256];
        memcpy(s,&evtdesc[7+firstsize+1], g);
        s[g]=0;

        writeDumpStr("DESC = ");
        writeDumpStr(s);
        writeDumpStr("\n");

        if (fs != "")
            fs+=", ";

        if (*s>=' ')
            fs+= s;
        else   
        if (*s>0)
            fs+= &s[1];
    }

    writeDumpStr("LONGDESC = ");
    writeDumpStr(fs);
    writeDumpStr("\n");

    data->setLongDescription(data->getLongDescription() + fs);   
}


void MMSEPGImportPlugin::parseDescription(char *desc, int len, MMSEPGData *data) {
	int i,tag,taglen;
	for (i=0;i<len;)
	{
		tag=*(desc+i) &0xff;
		taglen=*(desc+i+1) &0xff;

        writeDumpStr("TAG = " + iToStr(tag) + "\n");
        writeDumpStr("TAGLEN = " + iToStr(taglen) + "\n");

		if (taglen>0) {
			switch (tag) {
				case 0x4D: /* short evt desc, [title] [desc] */
					parseShortEventDesc(desc+i, data);
					break;
				case 0x4E: /* long evt descriptor */
					parseLongEventDesc(desc+i, data);
					break;
				default:
					break;
			}
		}
		i=i+taglen+2;
	}
}
  

void MMSEPGImportPlugin::throughBuffer(unsigned char *eitbuf, int len,
                                   vector<MMSTVChannelData *> channels, vector<MMSEPGData *> &epg) {
	long int    mjd;
	int         year,month,day,dll;
	eit_t       *e=(eit_t*)eitbuf;
	eit_event_t *evt=(eit_event_t*)((char*)eitbuf+EIT_LEN);
	
	/* check checksum */
    if( _dvb_crc32((uint8_t*)eitbuf, len) != 0) return;
    /* remove CRC */
    len=len-4; 
	
	/* for each event */
	while (evt!=NULL) {
        bool sidfound = false;

        /* check if service id can be found in channels */
        MMSTVChannelData *channel=NULL;
        int sid = HILO(e->service_id);
        for (unsigned int i = 0; i < channels.size(); i++) {
            if (channels.at(i)->getSID() == sid) {
                /* right, i can get the event */
                channel=channels.at(i);
                sidfound = true;
                break;
            }
        }

        /* length of message */
        if ((dll=HILO(evt->descriptors_loop_length))==0)
            /* zero, skip it */
            return;

        /* dump this event */
        writeDumpStr("\n\n>>>>>>>>>>>>\n\n");
        writeDump((void*)e, NULL, 1, EIT_LEN+EIT_EVENT_LEN+dll);
        writeDumpStr("\n\n------------\n\n");

        if (sidfound) {
            /* I need this service id */
            writeDumpStr("SERVICE ID = " + iToStr(sid) + "\n");

			mjd=HILO(evt->mjd);
			/*
			 *  * Use the routine specified in ETSI EN 300 468 V1.4.1,
			 *  * "Specification for Service Information in Digital Video Broadcasting"
			 *  * to convert from Modified Julian Date to Year, Month, Day.
			 *  */
			year = (int) ((mjd - 15078.2) / 365.25);
			month = (int) ((mjd - 14956.1 - (int) (year * 365.25)) / 30.6001);
			day = mjd - 14956 - (int) (year * 365.25) - (int) (month * 30.6001);
            int i;
			if (month == 14 || month == 15)
				i = 1;
			else
				i = 0;
			year += i ;
			month = month - 2 - i * 12;
		
            struct tm dvb_time;
			dvb_time.tm_mday = day;
			dvb_time.tm_mon = month;
			dvb_time.tm_year = year;
			dvb_time.tm_isdst = 0;
			dvb_time.tm_wday = dvb_time.tm_yday = 0;

            writeDumpStr("YEAR = " + iToStr(year) + "\n");
            writeDumpStr("MONTH = " + iToStr(month) + "\n");
            writeDumpStr("DAY = " + iToStr(day) + "\n");
		
			dvb_time.tm_sec =  bcdtoint(evt->start_time_s);
			dvb_time.tm_min =  bcdtoint(evt->start_time_m);
			dvb_time.tm_hour = bcdtoint(evt->start_time_h) + time_offset;

            writeDumpStr("HOUR = " + iToStr(bcdtoint(evt->start_time_h)) + "/" + iToStr(bcdtoint(evt->start_time_h) + time_offset) + "\n");
            writeDumpStr("MINUTE = " + iToStr(bcdtoint(evt->start_time_m)) + "\n");
            writeDumpStr("SECOND = " + iToStr(bcdtoint(evt->start_time_s)) + "\n");

			time_t start_time=mktime(&dvb_time);
	   
			dvb_time.tm_sec  += bcdtoint(evt->duration_s);
			dvb_time.tm_min  += bcdtoint(evt->duration_m);
			dvb_time.tm_hour += bcdtoint(evt->duration_h);

			time_t end_time=mktime(&dvb_time);
			
			/* bad date check */
            time_t now;
            time(&now);
            /* if the program ends one hour before, or one week from today -> ignore it */
            if ((difftime(end_time,now) >= -60) && (difftime(end_time,now)<=86400*7)) {
                /* good time */	
    			string sdate,stime;
    			getCurrentTimeString(NULL, &sdate, &stime, &start_time);
                string edate,etime;
    			getCurrentTimeString(NULL, &edate, &etime, &end_time);

                writeDumpStr("SDATE = " + sdate + "\n");
                writeDumpStr("STIME = " + stime + "\n");
                writeDumpStr("EDATE = " + edate + "\n");
                writeDumpStr("ETIME = " + etime + "\n");
    
                /* create new epg entry */
                MMSEPGData *data = new MMSEPGData;
                data->setChannel(channel); 
                data->setStartDate(sdate);
                data->setStartTime(stime);
                data->setEndDate(edate);
                data->setEndTime(etime);
   		
                writeDump((char*)evt+EIT_EVENT_LEN, NULL, 1, dll);

                parseDescription((char*)evt+EIT_EVENT_LEN, dll, data);

                
                epg.push_back(data);
            }
        }

        /* print separator */
        writeDumpStr("\n\n<<<<<<<<<<<<\n\n");

		/* skip to next entry */
        e=(eit_t*)((char*)e+EIT_EVENT_LEN+dll);
		evt=(eit_event_t*)((char*)evt+EIT_EVENT_LEN+dll);
		len=len-EIT_EVENT_LEN-dll;
		if (len<EIT_EVENT_LEN) return;
	}
}


bool MMSEPGImportPlugin::readEPG(vector<MMSTVChannelData *> channels, vector<MMSEPGData *> &epg,
                             string file, unsigned int timeout) {
    struct dmx_sct_filter_params sctFilterParams;
    unsigned char   buf[4096];
    int             fd_epg;
    bool            firsttime=true;

    /* open stream */	
    if ((fd_epg = open(file.c_str(), O_RDWR)) < 0) {
        logger.writeLog("Failed to open " + file);
        return false; 
    }

    if (this->dump) {
        /* open dump file */
        this->dumpfile = new MMSFile("epg.dump", MMSFM_APPEND);
        if (this->dumpfile->getLastError()) {
            logger.writeLog("Failed to open epg.dump");
            return false; 
        }
    }

    /* init filter */	
    memset(&sctFilterParams, 0, sizeof(sctFilterParams));
    sctFilterParams.pid = 18; /* EIT Data */
    sctFilterParams.timeout = 0;
    sctFilterParams.flags = DMX_IMMEDIATE_START;
    sctFilterParams.filter.filter[0] = 0;
    sctFilterParams.filter.mask[0] = 0;  

    /* calculate exit time */
    time_t to = time(NULL) + timeout;

    /* lets start */
    if (ioctl(fd_epg, DMX_SET_FILTER, &sctFilterParams) >= 0) {
        struct pollfd ufd;
    	memset(&ufd,0,sizeof(ufd));
    	ufd.fd=fd_epg;
    	ufd.events=POLLIN;
    	if (poll(&ufd,1,1000) > 0) {
            int cnt;
            /* until timeout or 0 bytes read */
    		while ((cnt = read(fd_epg, buf, sizeof(buf)))) {
                firsttime=false;
    			if (buf[0]==0)
    				continue;
                buf[cnt]=0;
    
                /* parse buffer */
    			throughBuffer(buf, cnt, channels, epg);
    			buf[0]=0;
    
                /* check for timeout */
    		  	if (time(NULL) > to)
                    break;
    	  	}
    	}
    }
    close(fd_epg);

    if (this->dumpfile) {
        delete this->dumpfile;
        this->dumpfile = NULL;
    }
    
    if (firsttime)
        return false;
    else
        return true;
}

void MMSEPGImportPlugin::writeDump(void *ptr, size_t *ritems, size_t size, size_t nitems) {
    if (this->dumpfile) {
        /* write buffer */
        if (!this->dumpfile->writeBuffer(ptr, ritems, size, nitems)) {
            /* error occured, stop writing */
            delete this->dumpfile;
            this->dumpfile = NULL;
        }
    }
}

void MMSEPGImportPlugin::writeDumpStr(string str) {
    writeDump((void *)(str.c_str()), NULL, 1, str.size());
}

