/***************************************************************************
 *   Copyright (C) 2005-2006 by                                            *
 *                                                                         *
 *      Stefan Schwarzer <sxs@morphine.tv>                                 *
 *      Guido Madaus     <bere@morphine.tv>                                *
 *      Jens Schneider   <pupeider@morphine.tv>                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License.        *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "mmsmusicplugin.h"

MMS_EXPORT_CENTRAL_PLUGIN(MMSMusicPlugin);

MMSMusicPlugin::MMSMusicPlugin() {
    this->srcdir    = "$HOME";
    this->mlpos     = -1;
    this->mmssound  = NULL;
    this->fireStart = false;
    this->fireStop  = false;
}

MMSMusicPlugin::~MMSMusicPlugin() {
}

bool MMSMusicPlugin::initialize(MMSPluginData data) {
    MMSConfigData   *config;
    DataSource      datasource;

    try {
        logger.setIdentity("MMSMusicPlugin");
    
        /* build data source */
        config = new MMSConfigData();
        datasource.setFilename(config->getDataDBName());
        delete config;
        
        /* save my plugin data */
        this->plugindata = data;
    
        logger.writeLog("initializing...");

        /* through the plugin properties */
        vector<MMSPropertyData*> probs = this->plugindata.getProperties();
        for (unsigned int i = 0; i < probs.size(); i++) {
            if (probs.at(i)->getParameter()=="SRCDIR") {
                /* source dir found */
                this->srcdir = probs.at(i)->getValue();
            }
        }

        /* load local theme */
        this->theme = themeManager->loadLocalTheme(this->plugindata.getPath());

        /* create windows */
        this->window = (MMSRootWindow*)dm.loadDialog(this->plugindata.getPath() + "/mmsmusicplugin.xml", this->theme);
        this->window->onHide->connect(sigc::mem_fun(this,&MMSMusicPlugin::onHide));
        this->menu = (MMSMenu *)dm["albummenu"];
        this->menu2 = (MMSMenu *)dm["playlist"];

        /* connect onSelectItem callback */
        this->menu->onSelectItem->connect(sigc::mem_fun(this,&MMSMusicPlugin::onSelectItem));

        logger.writeLog("initializing finished");
    
    } catch(MMSError *error) {
        logger.writeLog("Abort due to: " + error->getMessage());
    }
     
    return true;
}

bool MMSMusicPlugin::onEvent(IMMSEvent *event) {
    return false;
}

bool MMSMusicPlugin::shutdown() {
    if (this->mmssound)
        delete this->mmssound;
    return true;
}


bool MMSMusicPlugin::show() {
    /* return if already shown */
    if (this->window->isShown())
        return true;

    /* show the root window */
    this->window->show();

    /* have to try three times, waiting of blocked devices */
    int trycnt=0;
    while (!this->mmssound) {
        trycnt++;
        sleep(1);
        try {
#ifdef USE_MMSSOUND
            /* create MMSSound instance */
            this->mmssound = new MMSSound();
#else
            /* create MMSVideo instance */
            this->mmssound = new MMSVideo();
#endif

            /* create watcher thread */
            mt=new MMSMusicThread(this->mmssound);

            /* clear menu */
            this->menu->clear();

            /* fill album menu */
            string searchstring = MMSFILESEARCH_DEEPESTDIRENTRY_OF_FILE;
            searchstring+= "*.mp3";
            MMSFileSearch   *fs;
            logger.writeLog("Searching under: " + this->srcdir + " (" + substituteEnvVars(this->srcdir) + ")");
            if ((fs = new MMSFileSearch(substituteEnvVars(this->srcdir), searchstring))) {
//this->menu->setFocus(true);
                MMSFILEENTRY_LIST albumlist = fs->execute();
                for (MMSFILEENTRY_LIST::const_iterator i = albumlist.begin(); i != albumlist.end(); ++i) {
                    logger.writeLog("Found dir: " + (*i)->basename);
                    /* create menu item */
                    MMSButton *button;
                    button = (MMSButton *)menu->newItem();
                    ((MMSLabel*)(*button)["itemtext"])->setText((*i)->basename);
                    button->setData((*i)->name);
                    delete *i;
                }
                this->menu->setFocus(true);
                delete fs;
            }

            break;
        } catch(MMSError *error) {
            logger.writeLog("Abort due to: " + error->getMessage());
            if (trycnt>=3) {
                /* give up, hide the root window */
                this->window->hide();
                return false;
            }
        }
    }

    return true;
}

void MMSMusicPlugin::playNext() {

    if (this->mmssound) {
        if (this->mlpos < 0) {
            /* first time in the current playlist */
            if (this->musiclist.size() > 0)
                /* set start pos */
                this->mlpos = 0;
            else
                /* no entries */
                return;
        }
        else {
            /* reset length of the current item */
            char tb[32];
            sprintf(tb, "%02d:%02d", ((int)this->musiclist.at(this->mlpos)->length)/60, ((int)this->musiclist.at(this->mlpos)->length)%60);
            ((MMSLabel*)(*(this->menu2->getSelectedItem()))["length"])->setText(tb);

            /* switch to next entry */
            this->mlpos++;
            if (this->mlpos >= (int)this->musiclist.size())
                /* start from beginning */
                this->mlpos=0;
        }

        /* open new source */
        logger.writeLog("Start playing: " + this->musiclist.at(this->mlpos)->name);
        this->mmssound->openSource(this->musiclist.at(this->mlpos)->name);
        this->mmssound->play();
        this->menu2->setSelected(this->mlpos);
        /* start watcher */
        mt->setVal(this->musiclist.at(this->mlpos)->length,
                    (MMSLabel*)(*(this->menu2->getSelectedItem()))["length"]);
        mt->start();
    }
}

void MMSMusicPlugin::onFirstStart(MMSMusicThread *thread) {
    this->fireStart=true;
    playNext();
    this->fireStart=false;
}

void MMSMusicPlugin::onStop(MMSMusicThread *thread) {
    this->fireStop=true;
    playNext();
    this->fireStop=false;
}

void MMSMusicPlugin::onHide(MMSWindow *window, bool goback) {
    if (this->window == window) {
        /* root window */
        if (this->mt) {
            mt->onStop.clear();
        }
        if (this->mmssound) {
            this->mmssound->closeSource();
            sleep(1);
        }
        if (this->mt) {
            delete this->mt;
            this->mt = NULL;
        }
        if (this->mmssound) {
            delete this->mmssound;
            this->mmssound = NULL;
        }
    }
}

void MMSMusicPlugin::onSelectItem(MMSWidget *widget) {
    MMSButton       *button = (MMSButton *) widget;
    MMSFileSearch   *fs;

    /* wait of end of start */
    while (this->fireStart) msleep(100);

    /* stop playing, do not call onStop callback */
    if (this->mt) {
        mt->setFirstStart(false);
        mt->onFirstStart.clear();
        mt->onStop.clear();
    }
    if (this->mmssound) {
        this->mmssound->closeSource();
    }

    /* re-init onStop callback */
    mt->onFirstStart.connect(sigc::mem_fun(this,&MMSMusicPlugin::onFirstStart));
    mt->onStop.connect(sigc::mem_fun(this,&MMSMusicPlugin::onStop));

    /* clear menu */
    this->menu2->clear();

    /* clear music list */
    for(int i = (int)musiclist.size() - 1; i >= 0 ; i--) {
        delete musiclist.at(i);
        musiclist.erase(musiclist.end()-1);
    }

    /* search for files and create new music list and menu */
    if ((fs = new MMSFileSearch(button->getData(), "*.mp3"))) {
        MMSFILEENTRY_LIST filelist = fs->execute();
        for (MMSFILEENTRY_LIST::const_iterator i = filelist.begin(); i != filelist.end(); ++i) {
            logger.writeLog("Found file: " + (*i)->name);
            /* create menu item */
            MMSButton *button;
            button = (MMSButton *)menu2->newItem();
            logger.writeLog("Reading information from: " + (*i)->name);
            msleep(50);
            this->mmssound->openSource((*i)->name);

            MUSICINFO *musicinfo = new MUSICINFO;
            musicinfo->name = (*i)->name;
#ifdef USE_MMSSOUND
            musicinfo->length = this->mmssound->getTrackLength();
#else
            musicinfo->length = this->mmssound->getLength();
#endif
            logger.writeLog(" LENGTH: " + fToStr(musicinfo->length));
            this->musiclist.push_back(musicinfo);

            this->mmssound->closeSource();

            ((MMSLabel*)(*button)["filename"])->setText((*i)->basename);

            char tb[32];
            sprintf(tb, "%02d:%02d", ((int)musicinfo->length)/60, ((int)musicinfo->length)%60);
            ((MMSLabel*)(*button)["length"])->setText(tb);
            
            delete *i;
        }

        delete fs;

        /* mark as new loaded */
        this->mlpos = -1;
    }

    /* start watcher */
    while (mt->isRunning()) msleep(100);
    mt->setFirstStart(true);
    mt->start();
}


