/***************************************************************************
 *   Copyright (C) 2005-2006 by                                            *
 *                                                                         *
 *      Stefan Schwarzer <sxs@morphine.tv>                                 *
 *      Guido Madaus     <bere@morphine.tv>                                *
 *      Jens Schneider   <pupeider@morphine.tv>                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License.        *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "mmstvplugin.h"
#include "mmstvchannelservice.h"
#include "mmsepgservice.h"

MMS_EXPORT_CENTRAL_PLUGIN(MMSTVPlugin);

MMSTVPlugin::MMSTVPlugin() {
    currChannel = 0;
    currChannelName = "";
    this->mmstv = NULL;
}

MMSTVPlugin::~MMSTVPlugin() {
}

bool MMSTVPlugin::initialize(MMSPluginData data) {
    MMSConfigData   *config;
    DataSource      datasource;

    try {
        logger.setIdentity("MMSTVPlugin");
    
        /* build data source */
        config = new MMSConfigData();
        datasource.setFilename(config->getDataDBName());
        delete config;
        
        /* save my plugin data */
        this->plugindata = data;
    
        logger.writeLog("initializing...");

        /* get access to tvchannel service */
        tcs = new MMSTVChannelService(datasource);

        /* get access to epg service */
        eps = new MMSEPGService(datasource);

        /* create windows */
        this->window = new MMSRootWindow("","100%","576px",MMSALIGNMENT_CENTER,(MMSWINDOW_FLAGS)(MMSW_VIDEO|MMSW_INTERLACE));
        this->window->onHide->connect(sigc::mem_fun(this,&MMSTVPlugin::onHide));
        this->window->onHandleInput->connect(sigc::mem_fun(this,&MMSTVPlugin::onHandleInput));

        this->cmenu_window = (MMSMainWindow*)cmenu_dm.loadDialog(this->plugindata.getPath() + "/mmstvplugin_cmenu.xml");
        this->cmenu_window->onHide->connect(sigc::mem_fun(this,&MMSTVPlugin::onHide));
        this->cmenu = (MMSMenu *)cmenu_dm["channels"];
        this->cepgtext = (MMSTextBox*)(*cmenu_dm["cepgtext"])["textbox"];

        this->emenu_window = (MMSMainWindow*)emenu_dm.loadDialog(this->plugindata.getPath() + "/mmstvplugin_emenu.xml");
        this->emenu_window->onHide->connect(sigc::mem_fun(this,&MMSTVPlugin::onHide));
        this->ecaption = (MMSLabel *)emenu_dm["caption"];
        this->emenu = (MMSMenu *)emenu_dm["epg"];
        this->eepgtext = (MMSTextBox*)(*emenu_dm["eepgtext"])["textbox"];

        /* connect onSelectItem callback to the menu widgets */
        cmenu->onSelectItem->connect(sigc::mem_fun(this,&MMSTVPlugin::onSelectCItem));
        emenu->onSelectItem->connect(sigc::mem_fun(this,&MMSTVPlugin::onSelectEItem));

        /* connect onReturn callback to the menu widget */
        cmenu->onReturn->connect(sigc::mem_fun(this,&MMSTVPlugin::onReturn));

        logger.writeLog("initializing finished");
    
    } catch(MMSError *error) {
        logger.writeLog("Abort due to: " + error->getMessage());
    }
     
    return true;
}

bool MMSTVPlugin::onEvent(IMMSEvent *event) {
    return false;
}

bool MMSTVPlugin::shutdown() {
    if (this->mmstv)
        delete this->mmstv;
    delete tcs;
    delete eps;
    return true;
}


bool MMSTVPlugin::show() {
    /* return if already shown */
    if (this->window->isShown())
        return true;

    /* show the root window */
    this->window->show();

    /* have to try three times, waiting of blocked devices */
    int trycnt=0;
    while (!this->mmstv) {
        trycnt++;
        sleep(1);
        try {
            /* create MMSTV instance */
            this->mmstv = new MMSTV(this->window);
            break;
        } catch(MMSError *error) {
            logger.writeLog("Abort due to: " + error->getMessage());
            if (trycnt>=3) {
                /* give up, hide the root window */
                this->window->hide();
                return false;
            }
        }
    }

    if (channels.size()>0) {
        /* delete channels */
        for(unsigned int i = 0; i < channels.size();i++) {
            delete channels.at(i);
        }

        /* clear menu */
        cmenu->clear();
    }
    
    /* get all channels */
    channels = tcs->getAllChannels();

    /* fill menu */
    currChannel = 0;
    for(unsigned int i = 0; i < channels.size();i++) {
        /* create new menu item */          
        MMSButton *button;
        button = (MMSButton *)cmenu->newItem();
        ((MMSLabel*)(*button)["itemtext"])->setText(channels.at(i)->getName());

        /* get current epg data */
        MMSEPGData *cepg = eps->getCurrentEPGDataByChannel(channels.at(i));
        if (cepg) {
            /* process epg data */
            button->setData(buildEPGString(cepg));
            delete cepg;
        }

        if (channels.at(i)->getName() == currChannelName) {
            /* last channel found */
            currChannel = i;
        }
    }

    if (channels.size()>0) {
        /* select channel */
        this->cmenu->setSelected(currChannel);
    
        /* open channel */
        currChannelName = channels.at(currChannel)->getName();
        logger.writeLog("open channel: " + currChannelName);
        if (this->mmstv) {
            this->mmstv->openChannel(currChannelName);
        }
    }

    /* play tv stream */
    if (this->mmstv) {
        if (currChannelName!="")
            this->mmstv->play();
    }

    return true;
}

void MMSTVPlugin::onHide(MMSWindow *window, bool goback) {
    if (this->window == window) {
        /* root window */
        if (this->mmstv) {
            delete this->mmstv;
            this->mmstv = NULL;
        }
    }
    else
    if (this->cmenu_window == window) {
        /* channel window */
        if (goback) {
            /* switch back to old channel */
            this->cmenu->setSelected(currChannel);
        }
    }
    else
    if (this->emenu_window == window) {
        /* epg window */
    }
}

void MMSTVPlugin::onHandleInput(MMSWindow *window, DFBInputDeviceKeySymbol key) {
    unsigned int sel;

    switch(key) {
        case DIKS_CURSOR_DOWN:
        case DIKS_CURSOR_UP:
            /* reload EPG data */
            for(unsigned int i = 0; i < channels.size();i++) {
                /* get menu item */          
                MMSButton *button;
                button = (MMSButton *)cmenu->getItem(i);
        
                /* get current epg data */
                MMSEPGData *cepg = eps->getCurrentEPGDataByChannel(channels.at(i));
                if (cepg) {
                    /* process epg data */
                    button->setData(buildEPGString(cepg));
                    delete cepg;
                }
            }

            /* update textbox for selected item */
            MMSButton *button;
            button = (MMSButton *)cmenu->getSelectedItem();
            cepgtext->setText(button->getData());

            /* show channel menu */
            this->cmenu_window->show();
            break;
        case DIKS_INSERT:
            sel = this->cmenu->getSelected();
            if (sel > 0) {
                sel--;
                this->cmenu->setSelected(sel);
                switchChannel();
            }
            break;
        case DIKS_DELETE:
            sel = this->cmenu->getSelected();
            if (sel+1 < this->cmenu->getSize()) {
                sel++;
                this->cmenu->setSelected(sel);
                switchChannel();
            }
            break;
        case DIKS_CURSOR_LEFT:
        case DIKS_CURSOR_RIGHT:
            /* set caption */
            ecaption->setText(channels.at(this->cmenu->getSelected())->getName());
            /* reload EPG data */
            emenu->clear();
            vector<MMSEPGData *> epg = eps->getNextEPGDataByChannel(channels.at(this->cmenu->getSelected()));
            for(unsigned int i = 0; i < epg.size();i++) {
                /* create new menu item */          
                MMSButton *button;
                button = (MMSButton *)emenu->newItem();
                ((MMSLabel*)(*button)["itemtext"])->setText(epg.at(i)->getTitle());
                /* get values */
                string title = epg.at(i)->getTitle(); 
                string desc  = epg.at(i)->getDescription();
                string ldesc = epg.at(i)->getLongDescription();
                /* build epg info */
                string s="";
                s+= epg.at(i)->getStartDate() + ", " + epg.at(i)->getStartTime().substr(0,5);
                s+= " - " + epg.at(i)->getEndTime().substr(0,5) + "\n\n";
                s+= title;
                if ((title != desc)&&(desc != ""))
                    s+= "\n\n" + desc;
                if (ldesc != "")
                    s+= "\n\n" + ldesc;
                button->setData(s);
                delete epg.at(i);
            }
            this->emenu->setSelected(0);

            /* show epg menu */
            this->emenu_window->show();
            break;
    }
}

void MMSTVPlugin::onSelectCItem(MMSWidget *widget) {
    MMSButton *button = (MMSButton *) widget;
    cepgtext->setText(button->getData());
}

void MMSTVPlugin::onSelectEItem(MMSWidget *widget) {
    MMSButton *button = (MMSButton *) widget;
    eepgtext->setText(button->getData());
}

void MMSTVPlugin::onReturn(MMSWidget *widget) {
    switchChannel();
    this->cmenu_window->hide();
}

void MMSTVPlugin::switchChannel() {
    /* open channel */
    currChannel = this->cmenu->getSelected();
    MMSButton *button = (MMSButton *) this->cmenu->getItem(currChannel);
    currChannelName = ((MMSLabel*)(*button)["itemtext"])->getText();
    logger.writeLog("open channel: " + currChannelName);
    if (this->mmstv) {
        this->mmstv->openChannel(currChannelName);
        this->mmstv->play();
    }
}

string MMSTVPlugin::buildEPGString(MMSEPGData *epgdata) {
    if (!epgdata) return "";
    /* get values */
    string title = epgdata->getTitle(); 
    string desc  = epgdata->getDescription();
    string ldesc = epgdata->getLongDescription();
    /* build epg info */
    string s="";
    s+= epgdata->getStartTime().substr(0,5) + " - " + epgdata->getEndTime().substr(0,5);
    s+= "\n\n" + title;
    if ((title != desc)&&(desc != ""))
        s+= "\n\n" + desc;
    if (ldesc != "")
        s+= "\n\n" + ldesc;
    return s;
}


